import numpy as np

from .exceptions import UnphysicalValueException, \
    UnknownOrUnstableNuclideException, \
    NoDataException
from .database import ReadOnlyDatabase
from .decorators import asarray
from .inventory import UnstablesInventory

linspace = np.linspace
logspace = np.logspace

class EnergyGrid(object):
    """
        Basically a numpy array with some units
        and some protection on negative values
    """

    __slots__ = [ 'bounds' ]

    def __init__(self, bounds=linspace(0.0, 10e6, 10000)):
        """
            Energies in eV

            TODO: use a library to handle units
        """

        if np.any(bounds < 0):
            raise UnphysicalValueException("Energies cannot be negative.")

        self.bounds = bounds

    def __len__(self):
        return len(self.bounds)

    def __getitem__(self, i: int):
        return self.bounds[i]

    def __str__(self):
        return str(self.bounds)

    @property
    def nrofbins(self):
        return len(self)-1

    @property
    def units(self):
        """
            TODO: support other units
        """
        return "eV"

    @property
    @asarray
    def midpoints(self):
        """
            In eV
        """
        return [ (lower + self.bounds[i+1])*0.5 for i, lower in enumerate(self.bounds[:-1])]

    @property
    def minEnergy(self):
        """
            In eV
        """
        return np.min(self.bounds)

    @property
    def maxEnergy(self):
        """
            In eV
        """
        return np.max(self.bounds)

class LineComputor(object):
    """
        Needs testing!
    """

    __slots__ = [ 'db', 'grid', 'lines', 'values', 'linetype' ]

    def __init__(self, db: ReadOnlyDatabase, grid: EnergyGrid, type: str="gamma"):
        self.db = db

        self.grid = grid

        # here we just store lines and intensities
        # we loose knowledge of original nuclide
        # maybe later we will need this
        # TODO: better datastructure - ordereddict might be better
        self.lines = []

        # NOTE: values are intensities multiplied by the activity of each nuclide here
        self.values = []

        self.linetype = type

    def __call__(self, inventory: UnstablesInventory, *args, **kwargs):
        """
            Gets the lines from the full inventory

            throws an exception if nuclide is stable or is not in database
        """
        for zai, activity in inventory:
            name = self.db.getname(zai)
            # check it exists in database
            if name not in self.db:
                raise UnknownOrUnstableNuclideException(
                    "{} not in database - maybe too exotic or is it stable?".format(zai))

            # check that data exists for that decay type
            if self.linetype not in self.db.gettypes(name):
                raise NoDataException(
                    "{} does not have {} decay mode".format(name, self.linetype))

            self.lines.extend(self.db.getenergies(name, type=self.linetype))
            self.values.extend(self.db.getintensities(name, type=self.linetype)*activity)

        hist = [0.0]*self.grid.nrofbins

        if len(self.lines) > 0:
            # sort the lines in ascending energy to make it easier to bin in a histogram
            sorteddata = sorted(zip(self.lines, self.values), key=lambda pair: pair[0])
            self.lines = [x for x, _ in sorteddata]
            self.values = [y for _, y in sorteddata]

            # loop over lines to find appropriate bin
            ibin = 0
            for i, line in enumerate(self.lines):
                # loop over bounds from the last found bin
                for j in range(ibin, len(self.grid)-1):
                    lowerEnergy = self.grid[j]
                    upperEnergy = self.grid[j+1]
                    if line >= lowerEnergy and line < upperEnergy:
                        hist[j] += self.values[i]
                        ibin = j


        # should be density=False since we can have multiple lines per decay - cannot get numpy histogram to work
        # with the data correctly
        # return np.histogram(self.lines, bins=self.grid.bounds, weights=self.values, density=False)
        return hist, self.grid.bounds


class ActiGammaException(Exception):
    pass

class UnphysicalValueException(ActiGammaException):
    pass

class UnknownOrUnstableNuclideException(ActiGammaException):
    pass

class NoDataException(ActiGammaException):
    pass


import os
import json


from .decorators import asarray

# hacky but will do, the database is one large JSON file with line data
# we load it into a static data structure which is our database
__RAW_DATABASE_DECAY_2012_FILE__ = os.path.join(os.path.dirname(os.path.abspath(__file__)),
    'data', 'lines_decay_2012.min.json')

# potential to add other libraries here

class DatabaseJSONFileLoader(object):
    """
        Context manager to handle JSON datafile
    """
    def __init__(self, datafile=__RAW_DATABASE_DECAY_2012_FILE__):
        self.filename = datafile

    def __enter__(self):
        with open(self.filename, 'rt') as fjson:
            return json.loads(fjson.read())

    def __exit__(self, *args):
        pass

# a facade layer to interact with database
class ReadOnlyDatabase(object):
    """
        TODO: handle uncertainties too
    """
    def __init__(self, datasource):
        self._raw = {}
        with datasource as db:
            self._raw = db

    def __contains__(self, nuclide: str) -> bool:
        return nuclide in self._raw

    @property
    @asarray
    def allnuclides(self) -> [str]:
        return [k for k, _ in self._raw.items()]

    def allnuclidesoftype(self, type: str="gamma") -> [str]:
        return [k for k, _ in self._raw.items() if type in self._raw[k].keys()]

    def gettypes(self, nuclide: str) -> [str]:
        """
            Check if it has that particular decay type
        """
        return [k for k, _ in self._raw[nuclide].items() if k not in ["zai", "halflife"] ]

    def hastype(self, nuclide: str, type: str="gamma") -> bool:
        """
            Check if it has that particular decay type
        """
        return type in self._raw[nuclide]

    def getname(self, zai: int) -> str:
        """
            ZAI
        """
        for k, v in self._raw.items():
            if v['zai'] == zai:
                return k
        return None

    def getzai(self, nuclide: str) ->int:
        """
            ZAI
        """
        return self._raw[nuclide]['zai']

    def gethalflife(self, nuclide: str):
        """
            Half-life in seconds
        """
        return self._raw[nuclide]['halflife']

    @asarray
    def getenergies(self, nuclide: str, type: str="gamma"):
        """
            Defaults to gamma lines
        """
        return self._raw[nuclide][type]['lines']['energies']

    @asarray
    def getintensities(self, nuclide: str, type: str="gamma"):
        """
            Defaults to gamma lines

            Also multiplies by normalisation constant
        """
        return [ intensity*self._raw[nuclide][type]['lines']['norms'][i] 
                for i,intensity in enumerate(self._raw[nuclide][type]['lines']['intensities'])]